/*
   Ray Tracing Point Lights Demo.
   Created by the Programming Ace.
   www.UltimateGameProgramming.com

   The purpose of this demo is to build off of the previous ray tracing demos
   and add point lights to the scene.  Point lights are just like our old lights
   except that we must add one small calculation called attenuation.  This
   value causes the light to decrease over distance like you would expect
   from a point light source.  This way instead of having a high intensity,
   infinite light source like before, we can now have a finite point light.
   To calculate attenuation we will be using this equation:
   
   attenuation = 1 / ((d * d) / (lights[l].radius * lights[l].radius));
   
   where d = distance (length(intersect point - light pos))
   
   Once you have this value you simply multiple it by the light contribution
   like so
   
   final color = attenuation * light color * diffuse (light coef) * object color;
   
   In the past we did this in our lighting calculation:

   final color = light color * diffuse (light coef) * object color;
   
   Also note that in this demo I've added a light type and radius to the
   light class.
*/


// Demo: Ray Tracing Point Lights.


// All the necessary includes...
#include<iostream>
#include"Shape.h"
#include"RayTracer.h"
#include"Light.h"
#include"TgaImage.h"

using namespace std;

// Width/Height resolution and total number of objects.
#define WIDTH_RES 640
#define HEIGHT_RES 480
#define TOTAL_SHAPES 15
#define TOTAL_LIGHTS 1


// Function prototype...
bool InitializeRT();
void ClearBuffer(unsigned char *buffer, int size, int r, int g, int b);
void Shutdown();


// Primary render array and the look up array of directions.
unsigned char *primaryBuffer = 0;
CVector4 *directionList = 0;

// List of objects, light source, and tga object to save the rendered scene.
Shape shapes[TOTAL_SHAPES];
CLight lights[TOTAL_LIGHTS];
Ray primaryRay;


int main()
{
   cout << "Ultimate Game Programming Ray Tracer..." << endl;
   cout << "Created by Allen Sherrod..." << endl << endl;

   // Initialize all arrays and data nescessary.
   if(!InitializeRT())
      {
         cout << "Initialization failed!  Exiting." << endl;
         return 0;
      }

   cout << "Clearing primary buffer..." << endl;

   // Clear the rendered scene to a default color (I choose black).
   ClearBuffer(primaryBuffer, WIDTH_RES * HEIGHT_RES * 3, 0, 0, 0);
   
   cout << "Primary buffer cleared and ready to be drawn to..." << endl << endl;
   cout << "Starting trace." << endl;

   // Trace through the scene and create the rendered scene.
   int bufferIndex = 0;
   for(int y = 0; y < HEIGHT_RES; y++)
      {
         if(y == 0) cout << "  0% complete..." << endl;
         if(y == (int)(HEIGHT_RES * 0.4)) cout << " 25% complete..." << endl;
         if(y == (int)(HEIGHT_RES * 0.5)) cout << " 50% complete..." << endl;
         if(y == (int)(HEIGHT_RES * 0.75)) cout << " 75% complete..." << endl;
         if(y == (int)(HEIGHT_RES - 1)) cout << "100% complete..." << endl;

         for(int x = 0; x < WIDTH_RES; x++, bufferIndex+=3)
            {
               float r = 0, g = 0, b = 0;

               // Grab the direction for this current pixel.
               primaryRay.direction = directionList[x + y * WIDTH_RES];
               
               TraceScene(shapes, TOTAL_SHAPES, lights, TOTAL_LIGHTS,
                          primaryRay, r, g, b);

               // Convert and clamp to 0 to 255 from 0 to 1.
               r *= 255; g *= 255; b *= 255;
               if(r > 255) r = 255;
               if(g > 255) g = 255;
               if(b > 255) b = 255;

               // Save the color in our back buffer array.
               primaryBuffer[bufferIndex + 0] = (int)r;
               primaryBuffer[bufferIndex + 1] = (int)g;
               primaryBuffer[bufferIndex + 2] = (int)b;
            }
      }

   cout << endl;
   cout << "Trace over." << endl << endl;
   
   cout << "Preparing to save scene as RayTracedScene.tga..." << endl;
   
   // Take our ray traced scene and save it as a .tga image file.
   WriteTGA("RayTracedScene.tga", WIDTH_RES, HEIGHT_RES, primaryBuffer);

   cout << "Traced scene saved...shut down starting..." << endl << endl;

   // Release all resources and data we used.
   Shutdown();

   cout << "Exiting..." << endl;

   return 1;
}


bool InitializeRT()
{
   // Release all previous data.
   Shutdown();

   // Allocate rendering buffer.
	primaryBuffer = new unsigned char[WIDTH_RES * HEIGHT_RES * 3];

   // Error checking...
   if(primaryBuffer == NULL)
      return false;

	// Allocate the ray directions.
	directionList = GenerateRayDirections(WIDTH_RES, HEIGHT_RES, 255);

   // Error checking...
   if(directionList == NULL)
      return false;
   
   // Set the primary ray.
   primaryRay.origin.x = 0;
   primaryRay.origin.y = 0;
   primaryRay.origin.z = -500;

   // Setup the lights.
   lights[0] = CLight(POINT_LIGHT, CVector4(0, 300, 0), 1, 1, 1, 500);

   // Initialize the shapes we are using.
   shapes[0].type = SPHERE;
   shapes[0].sphere = new Sphere(-400, -200, 0, 100, 1, 1, 1, 1.0f, 0.0f);
   
   shapes[1].type = SPHERE;
   shapes[1].sphere = new Sphere(0, -200, 200, 100, 1, 1, 1, 1.0f, 0.0f);

   shapes[2].type = SPHERE;
   shapes[2].sphere = new Sphere(400, -200, 0, 100, 1, 1, 1, 1.0f, 0.0f);
   
   // Ground.
   shapes[3].type = TRIANGLE;
   shapes[3].triangle = new Triangle(CVector4(-600, -300, -800),
                                     CVector4(-600, -300, 800),
                                     CVector4(600, -300, 800), 1, 1, 1, 0, 0);
   shapes[4].type = TRIANGLE;
   shapes[4].triangle = new Triangle(CVector4(600, -300, 800),
                                     CVector4(600, -300, -800),
                                     CVector4(-600, -300, -800), 1, 1, 1, 0, 0);
   // Left.
   shapes[5].type = TRIANGLE;
   shapes[5].triangle = new Triangle(CVector4(-600, -300, 800),
                                     CVector4(-600, -300, -800),
                                     CVector4(-600, 550, -800), 1, 0, 0, 0, 0);
   shapes[6].type = TRIANGLE;
   shapes[6].triangle = new Triangle(CVector4(-600, 550, -800),
                                     CVector4(-600, 550, 800),
                                     CVector4(-600, -300, 800), 1, 0, 0, 0, 0);
   // Right.
   shapes[7].type = TRIANGLE;
   shapes[7].triangle = new Triangle(CVector4(600, -300, -800),
                                     CVector4(600, -300, 800),
                                     CVector4(600, 550, 800), 0, 0, 1, 0, 0);
   shapes[8].type = TRIANGLE;
   shapes[8].triangle = new Triangle(CVector4(600, 550, 800),
                                     CVector4(600, 550, -800),
                                     CVector4(600, -300, -800), 0, 0, 1, 0, 0);
   // Roof.
   shapes[9].type = TRIANGLE;
   shapes[9].triangle = new Triangle(CVector4(-600, 550, 800),
                                     CVector4(-600, 550, -800),
                                     CVector4(600, 550, -800), 1, 1, 1, 0, 0);
   shapes[10].type = TRIANGLE;
   shapes[10].triangle = new Triangle(CVector4(600, 550, -800),
                                     CVector4(600, 550, 800),
                                     CVector4(-600, 550, 800), 1, 1, 1, 0, 0);
   // Front.
   shapes[11].type = TRIANGLE;
   shapes[11].triangle = new Triangle(CVector4(-600, -300, -800),
                                     CVector4(600, -300, -800),
                                     CVector4(600, 550, -800), 1, 1, 1, 0, 0);
   shapes[12].type = TRIANGLE;
   shapes[12].triangle = new Triangle(CVector4(600, 550, -800),
                                     CVector4(-600, 550, -800),
                                     CVector4(-600, -300, -800), 1, 1, 1, 0, 0);
   // Back.
   shapes[13].type = TRIANGLE;
   shapes[13].triangle = new Triangle(CVector4(600, -300, 800),
                                     CVector4(-600, -300, 800),
                                     CVector4(-600, 550, 800), 0, 1, 0, 0, 0);
   shapes[14].type = TRIANGLE;
   shapes[14].triangle = new Triangle(CVector4(-600, 550, 800),
                                     CVector4(600, 550, 800),
                                     CVector4(600, -300, 800), 0, 1, 0, 0, 0);
   return true;
}


void ClearBuffer(unsigned char *buffer, int size, int r, int g, int b)
{
   // Loop through the entire array and set the buffer to the specified color.
   for(int p = 0; p < size; p += 3)
      {
         buffer[p + 0] = r; buffer[p + 1] = g; buffer[p + 2] = b;
      }
}


void Shutdown()
{
   // Release all resources we used.
   if(directionList != NULL) delete[] directionList;
   if(primaryBuffer != NULL) delete[] primaryBuffer;


   for(int i = 0; i < TOTAL_SHAPES; i++)
      {
         if(shapes[i].type == SPHERE && shapes[i].sphere != NULL)
            delete shapes[i].sphere;
         else if(shapes[i].type == TRIANGLE && shapes[i].triangle != NULL)
            delete shapes[i].triangle;
         else if(shapes[i].type == PLANE && shapes[i].plane != NULL)
            delete shapes[i].plane;
      }
}


// Recap:

// Adding point lights to the scene required one small additional calculation
// to our lighting equation.  With point lights our scene look much, much better.
// Adding spot light are also as simple but I'll leave that as a exercise for you.
// Play around with the light radius to see different results.


// Copyright February 2005
// All Rights Reserved!
// Allen Sherrod
// ProgrammingAce@UltimateGameProgramming.com
// www.UltimateGameProgramming.com